/*
 * Copyright (c) 2015-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Thu Oct  4 2018
 */

#ifndef H_GPS_CODELS
#define H_GPS_CODELS

#include <sys/socket.h>
#include <sys/types.h>

#include <aio.h>
#include <errno.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#include "gps_c_types.h"

/* communication channel */
struct gps_channel_s {
  char path[1024];
  uint32_t baud;
  int fd;

  uint8_t buf[2][512]; /* ring buffer with mirrorring for recv() */
  uint16_t r, w;
};

/* connected device */
struct gps_device_s {
  struct gps_channel_s channel;
  const struct gps_driver_s *driver;
  void *data;
};

struct gps_devicefd_s {
  const int fd;
};

/* hardware drivers */
enum gps_ctl_s {
  GPSCTL_BAUD,		/* uint32_t */
  GPSCTL_UNFIX,		/* void */
  GPSCTL_FIX,		/* gps_llh_s, gps_enu_s */
  GPSCTL_OUTRTCM,	/* uint16_t, double, double */
  GPSCTL_INRTCM,	/* uint8_t *, size_t */

  GPSCTL_Q_PARAM,	/* gps_param_s * */
  GPSCTL_Q_RTCM		/* gps_rtcm_stats_s * */
};

struct gps_driver_s {
  const char *protocol;

  void *(*match)(struct gps_channel_s *chan);
  int (*configure)(struct gps_device_s *dev);
  int (*unconfigure)(struct gps_device_s *dev);
  int (*decode)(struct gps_device_s *dev, gps_solution_s *solution,
                sequence_octet *rtcm_data);
  int (*ctl)(struct gps_device_s *dev, enum gps_ctl_s request, ...);
};


/* rtk server and client */
struct rtk_addrinfo_s {
  struct sockaddr_storage addr;
  size_t addrlen;
  struct timeval keepalive;		/* deadline for keepalive */
  struct rtk_addrinfo_s *next;
};

struct gps_rtk_conn_s {
  int fd4, fd6;				/* server/client sockets */
  struct rtk_addrinfo_s *clients;	/* clients to send RTK to */
  struct rtk_addrinfo_s *servers;	/* servers to recv RTK from */
};


#ifdef __cplusplus
extern "C" {
#endif

  int		gpstty_open(struct gps_channel_s *chan, const char *device);
  int		gpstty_setspeed(struct gps_channel_s *chan, uint32_t speed);
  ssize_t	gpstty_recv(struct gps_channel_s *chan,
                        const struct timeval *deadline);
  const void *	gpstty_data(struct gps_channel_s *chan, size_t advance,
                        size_t *len);
  int		gpstty_sendmsg(const struct gps_channel_s *chan,
                        const char *fmt, ...);
  int		gpstty_vsendmsg(const struct gps_channel_s *chan,
                        const char *fmt, va_list args);
  int		gpstty_send(const struct gps_channel_s *chan,
                        const void *buf, size_t n);
  void		gpstty_close(struct gps_channel_s *chan);

  or_time_ts	gpsconv_utcmktime(struct tm *tm, int32_t nano);
  or_time_ts	gpsconv_gpsutc(uint16_t week, uint32_t ms, int16_t leaps);
  or_time_ts	gpsconv_utcts(or_time_ts utc, struct timeval rts, double period,
                        double tdelay);
  void		gpsconv_setref(const gps_llh_s *llh);
  void		gpsconv_llhecef(const gps_llh_s *llh, gps_ecef_s *ecef);
  void		gpsconv_ecefllh(const gps_ecef_s *ecef, gps_llh_s *llh);
  void		gpsconv_ecefenu(const gps_ecef_s *ecef, gps_enu_s *enu);
  void		gpsconv_decefenu(const gps_ecef_s *ecef, gps_enu_s *enu);
  void		gpsconv_ecefenu_stddev(const gps_ecef_s *ecef_stddev,
                        gps_enu_s *enu_stddev);
  void		gpsconv_llhenu(const gps_llh_s *llh, gps_enu_s *enu);
  void		gpsconv_llhavg(const gps_llh_s *llh, const gps_enu_s *llh_sigma,
                        gps_llh_s *avg, gps_enu_s *avg_sigma);

  int		gpsrtk_conn_open(gps_rtk_conn_s *rtk_conn, uint16_t port);
  int		gpsrtk_server_open(const char host[64], uint16_t port,
                        gps_rtk_conn_s *rtk_conn);
  int		gpsrtk_recv(gps_rtk_conn_s *rtk_conn, sequence_octet *rtk_in);
  int		gpsrtk_send(gps_rtk_conn_s *rtk_conn,
                        sequence_octet *rtk_out);
  int		gpsrtk_keepalive(gps_rtk_conn_s *rtk_conn);
  void		gpsrtk_conn_close(gps_rtk_conn_s *rtk_conn);
  int		gpsrtk_is_rtcm(const uint8_t *raw, uint16_t *type,
                        size_t *length);
  int		gpsrtk_match_rtcm(const uint8_t *raw, size_t max,
                        const uint8_t **match, uint16_t *type, size_t *length);
  int		gpsrtk_copy_rtcm_stats(gps_rtcm_stats_s *dst,
                        gps_rtcm_stats_s *src);

#ifdef __cplusplus
}
#endif

/* RTCM stats helpers - these are macros for type independency */
#define rtcmidseq_insert(sequence, id, index, ...)                      \
  do {                                                                  \
    size_t index;                                                       \
    for (index = 0; index < (sequence)->_length; index++)               \
      if ((sequence)->_buffer[index].rtcmid == (id)) break;             \
                                                                        \
    if (index >= (sequence)->_length &&                                 \
        !genom_sequence_reserve(sequence, index + 1)) {                 \
      memset(                                                           \
        &(sequence)->_buffer[index], 0, sizeof(*(sequence)->_buffer));  \
      (sequence)->_buffer[index].rtcmid = (id);                         \
      (sequence)->_length = index + 1;                                  \
    }                                                                   \
    if (index >= (sequence)->_length) break;                            \
    __VA_ARGS__;                                                        \
  } while(0)

#define rtcmidseq_rm(sequence, id)                                      \
  do {                                                                  \
    size_t i;                                                           \
    for (i = 0; i < (sequence)->_length; i++)                           \
      if ((sequence)->_buffer[i].rtcmid == (id)) {                      \
        memmove(                                                        \
          &(sequence)->_buffer[i], &(sequence)->_buffer[i+1],           \
          ((sequence)->_length-i-1) * sizeof(*(sequence)->_buffer));    \
        (sequence)->_length--;                                          \
        break;                                                          \
      }                                                                 \
  } while(0)



/* e_sys helper */
static inline genom_event
gps_e_sys_error(const char *s, genom_context self)
{
  gps_e_sys_detail d;
  size_t l = 0;

  d.code = errno;
  if (s) {
    strncpy(d.what, s, sizeof(d.what) - 3);
    l = strlen(s);
    strcpy(d.what + l, ": ");
    l += 2;
  }
  strerror_r(d.code, d.what + l, sizeof(d.what) - l);
  return gps_e_sys(&d, self);
}


/* logging */
struct gps_log_s {
  struct aiocb req;
  char buffer[4096];
  bool pending, skipped;
  uint32_t decimation;
  size_t missed, total;

# define gps_logfmt	" %.17g "
# define gps_log_header_fmt                                             \
  "ts utc fix fix# sats "                                               \
  "lat lon h "                                                          \
  "peast pnorth pup veast vnorth vup "                                  \
  "pseast psnorth psup vseast vsnorth vsup"
# define gps_log_fmt                                                    \
  "%d.%09d %d.%09d %s %d %d "                                           \
  gps_logfmt gps_logfmt gps_logfmt                                      \
  gps_logfmt gps_logfmt gps_logfmt gps_logfmt gps_logfmt gps_logfmt     \
  gps_logfmt gps_logfmt gps_logfmt gps_logfmt gps_logfmt gps_logfmt
};

static inline const char *
gps_fixstr(gps_fix_s fix)
{
  switch(fix) {
    case gps_fix_none: return "none";
    case gps_fix_interp: return "interp";
    case gps_fix_single: return "single";
    case gps_fix_diff: return "diff";
    case gps_fix_rtk_float: return "rtk-float";
    case gps_fix_rtk: return "rtk";
    case gps_fix_fixed: return "fixed";
  }
  return "?";
}

#endif /* H_GPS_CODELS */
