/*
 * Copyright (c) 2018-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Oct 18 2018
 *                                Antonio Enrique Jimenez Cano on Thu Nov 2019
 */
#include "acgps.h"

#include <sys/time.h>

#include <err.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "codels.h"

static void *		ublox_match(struct gps_channel_s *chan);
static int		ublox_configure(struct gps_device_s *dev);
static int		ublox_unconfigure(struct gps_device_s *dev);
static int		ublox_decode(struct gps_device_s *dev,
                                struct gps_solution_s *solution,
                                sequence_octet *rtk_data);
static int		ublox_ctl(struct gps_device_s *dev,
                                enum gps_ctl_s request, ...);


const struct gps_driver_s ublox_driver = {
  .protocol = "u-blox",

  .match = ublox_match,
  .configure = ublox_configure,
  .unconfigure = ublox_unconfigure,
  .decode = ublox_decode,
  .ctl = ublox_ctl
};


/* --- UBX protocol -------------------------------------------------------- */

enum ublox_msgcid {
  mcid_ACK_ACK = 0x0501,
  mcid_ACK_NAK = 0x0500,

  mcid_CFG_DGNSS = 0x0670,
  mcid_CFG_INF = 0x0602,
  mcid_CFG_MSG = 0x0601,
  mcid_CFG_PMS = 0x0686,
  mcid_CFG_PRT = 0x0600,
  mcid_CFG_RATE = 0x0608,
  mcid_CFG_TMODE3 = 0x0671,

  mcid_INF_ERROR = 0x0400,
  mcid_INF_WARN = 0x0401,
  mcid_INF_NOTICE = 0x0402,
  mcid_INF_TEST = 0x0403,
  mcid_INF_DEBUG = 0x0404,

  mcid_MON_MSGPP = 0x0a06,
  mcid_MON_TXBUF = 0x0a08,
  mcid_MON_VER = 0x0a04,

  mcid_NAV_EOE = 0x0161,
  mcid_NAV_HPPPOSLLH = 0x0114,
  mcid_NAV_PVT = 0x0107,

  mcid_RXM_RTCM = 0x0232,

  mcid_TIM_VRFY = 0x0d06,

  mcid_NMEA_GGA = 0xf000,
  mcid_NMEA_GLL = 0xf001,
  mcid_NMEA_GSA = 0xf002,
  mcid_NMEA_GSV = 0xf003,
  mcid_NMEA_RMC = 0xf004,
  mcid_NMEA_VTG = 0xf005,

  mcid_RTCM_1005 = 0xf505,
  mcid_RTCM_1074 = 0xf54a,
  mcid_RTCM_1077 = 0xf54d,
  mcid_RTCM_1084 = 0xf554,
  mcid_RTCM_1087 = 0xf557,
  mcid_RTCM_1124 = 0xf57c,
  mcid_RTCM_1127 = 0xf57f,
  mcid_RTCM_1230 = 0xf5e6
};

enum ublox_portid {
  prtid_I2C =	0,
  prtid_UART1 =	1,
  prtid_USB =	3,
  prtid_SPI =	4,

  prtid_INVALID =	255
};

struct __attribute__((packed)) ublox_ack {
  uint8_t msgclass, msgid;
};

struct __attribute__((packed)) ublox_cfg_dgnss {
  uint8_t	dgnssMode; /* 2:RTK-float 3:RTK-fixed */
  uint8_t	res[3];
};

struct __attribute__((packed)) ublox_cfg_inf {
  uint8_t proto;
  uint8_t reserved[3];
  uint8_t mask[6];
};

struct __attribute__((packed)) ublox_cfg_msg {
  uint8_t	msgclass, msgid;
  uint8_t	rate[6];
};

struct __attribute__((packed)) ublox_cfg_pms {
  uint8_t	version;
  uint8_t	powersetupvalue;
  uint16_t	period;
  uint16_t	ontime;
  uint8_t	res[2];
};

struct __attribute__((packed)) ublox_cfg_prt {
  uint8_t	portid;
  uint8_t	res[1];
  uint16_t	txready;
  union {
    struct {
      uint32_t	mode;
      uint32_t	baud;
      uint16_t	inproto;
      uint16_t	outproto;
      uint16_t	flags;
      uint8_t	res[2];
    } uart;
    struct {
      uint8_t	res1[8];
      uint16_t	inproto;
      uint16_t	outproto;
      uint8_t	res2[4];
    } usb;
  };
};

struct __attribute__((packed)) ublox_cfg_rate {
  uint16_t	measrate;
  uint16_t	navrate;
  uint16_t	timeref;
};

struct __attribute__((packed)) ublox_cfg_tmode3 {
  uint8_t	version;
  uint8_t	res1[1];
  uint16_t	flags;
  union {
    struct { int32_t lat, lon, height; };
    struct { int32_t x, y, z; };
  };
  union {
    struct { int8_t lathp, lonhp, heighthp; };
    struct { int8_t xhp, yhp, zhp; };
  };
  uint8_t	res2[1];
  uint32_t	fixedposacc;
  uint32_t	svinmindur;
  uint32_t	svinacclimit;
  uint8_t	res3[8];
};

struct __attribute__((packed)) ublox_inf {
  char		str; /* variable size string */
};

struct __attribute__((packed)) ublox_mon_msgpp {
  uint16_t msg1[8];
  uint16_t msg2[8];
  uint16_t msg3[8];
  uint16_t msg4[8];
  uint16_t msg5[8];
  uint16_t msg6[8];
  uint32_t skipped[6];
};

struct __attribute__((packed)) ublox_mon_txbuf {
  uint16_t pending[6];
  uint8_t usage[6];
  uint8_t peakusage[6];
  uint8_t tusage;
  uint8_t tpeakusage;
  uint8_t errors;
  uint8_t res[1];
};

struct __attribute__((packed)) ublox_mon_ver {
  char swver[30];
  char hwver[10];
  struct { char data[30]; } ext[];
};

struct __attribute__((packed)) ublox_nav_eoe {
  uint32_t	itow;
};

struct __attribute__((packed)) ublox_nav_hppposllh {
  uint8_t	version;
  uint8_t	res[3];
  uint32_t	itow;
  int32_t	lon, lat, height, hmsl;
  int8_t	lonhp, lathp, heighthp, hmslhp;

  uint32_t	hacc, vacc;
};

struct __attribute__((packed)) ublox_nav_pvt {
  uint32_t	itow;
  uint16_t	year;
  uint8_t	month, day, hour, min, sec;
  uint8_t	valid;
  uint32_t	tacc;
  int32_t	nano;

  uint8_t	fixtype, flags, flags2;
  uint8_t	numsv;

  int32_t	lon, lat, height, hmsl;
  uint32_t	hacc, vacc;

  int32_t	veln, vele, veld;
  int32_t	gspeed;
  int32_t	headmot;
  uint32_t	sacc, headacc;

  uint16_t	pdop;
  uint8_t	res[6];

  int32_t	headveh;
  int16_t	magdec;
  uint16_t	magacc;
};

struct __attribute__((packed)) ublox_rxm_rtcm {
  uint8_t	version;
  uint8_t	flags;
  uint8_t	res[2];
  uint16_t	refstation;
  uint16_t	msgtype;
};


struct __attribute__((packed)) ublox_hdr {
  uint8_t	sync1, sync2;
  uint8_t	msgclass, msgid;
#define ublox_msgcid16(h) ((uint16_t)(((h)->msgclass << 8) | (h)->msgid))
  uint16_t	msglen;
  uint8_t	payload[];
};

struct __attribute__((packed)) ublox_msg {
  struct ublox_hdr hdr;
  union {
    struct ublox_ack ack;
    struct ublox_cfg_dgnss cfg_dgnss;
    struct ublox_cfg_inf cfg_inf;
    struct ublox_cfg_msg cfg_msg;
    struct ublox_cfg_pms cfg_pms;
    struct ublox_cfg_prt cfg_prt;
    struct ublox_cfg_rate cfg_rate;
    struct ublox_cfg_tmode3 cfg_tmode3;
    struct ublox_inf inf;
    struct ublox_mon_msgpp mon_msgpp;
    struct ublox_mon_txbuf mon_txbuf;
    struct ublox_mon_ver mon_ver;
    struct ublox_nav_eoe nav_eoe;
    struct ublox_nav_hppposllh nav_hppposllh;
    struct ublox_nav_pvt nav_pvt;
    struct ublox_rxm_rtcm rxm_rtcm;
  } data[];
};

#define ublox_msg(len)                                              \
  struct __attribute__((packed)) {                                  \
    struct ublox_msg	msg;                                        \
    uint8_t		payload[len];                               \
    uint16_t		cksum;                                      \
  }


/* --- local data ---------------------------------------------------------- */

struct ublox_driver_data_s {
  char model[32];		/* device model */
  double protver;		/* protocol version */
  uint8_t port;			/* current communication port */
  double period;		/* configured solution period */

  gps_rtcm_stats_s rtcmstats;	/* RTCM message counters */
  uint32_t skipped;		/* skipped bytes by device */
  double tdelay;		/* estimated transmission delay for NAV-* */

  /* last solution messages to be reassembled */
  struct ublox_nav_hppposllh hppposllh;
  struct ublox_nav_pvt pvt;
};

static void	ublox_solution(uint32_t itow, const struct timeval *tv,
                        struct ublox_driver_data_s *ddata,
                        struct gps_solution_s *solution);
static int	ublox_ack_msg(struct gps_channel_s *chan, struct ublox_msg *msg,
                        enum ublox_msgcid mcid, uint16_t len, int timeout);
static size_t	ublox_recv_msg(struct gps_channel_s *chan,
                        const struct timeval *deadline,
                        const struct ublox_msg **ubx,
                        const uint8_t **rtcm, uint16_t *rtcmid);
static int	ublox_send_msg(struct gps_channel_s *chan,
                        struct ublox_msg *msg, enum ublox_msgcid mcid,
                        uint16_t len);
static uint16_t	ublox_cksum(const struct ublox_hdr *hdr);


/* --- ublox_match --------------------------------------------------------- */

static void *
ublox_match(struct gps_channel_s *chan)
{
  struct ublox_driver_data_s *ddata;
  const struct ublox_msg *msg;
  struct timeval deadline;
  unsigned int i;
  int s;

  /* try for at most 500ms */
  gettimeofday(&deadline, NULL);
  deadline.tv_usec += 500 * 1000;
  if (deadline.tv_usec >= 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  /* send get version message */
  ublox_msg(0) ver;
  s = ublox_send_msg(chan, &ver.msg, mcid_MON_VER, 0);
  if (s) return NULL;

  /* receive */
  ddata = NULL;
  do {
    if (ublox_recv_msg(chan, &deadline, &msg, NULL, NULL) == 0)
      return NULL;

    /* match MON-VER message */
    if (ublox_msgcid16(&msg->hdr) != mcid_MON_VER) continue;

    /* match */
    ddata = calloc(1, sizeof(*ddata));
    if (!ddata) return NULL;

    warnx("hw: software version %s", msg->data->mon_ver.swver);
    warnx("hw: hardware version %s", msg->data->mon_ver.hwver);
    snprintf(ddata->model, sizeof(ddata->model),
             "u-blox %s", msg->data->mon_ver.hwver);
    for(i = 0; offsetof(struct ublox_mon_ver, ext[i]) < msg->hdr.msglen; i++) {
      warnx("hw: %s", msg->data->mon_ver.ext[i].data);
      if (!strncmp(msg->data->mon_ver.ext[i].data, "MOD=", 4))
        snprintf(ddata->model, sizeof(ddata->model),
                 "u-blox %s", msg->data->mon_ver.ext[i].data + 4);
      if (!strncmp(msg->data->mon_ver.ext[i].data, "PROTVER=", 8))
        ddata->protver = strtod(msg->data->mon_ver.ext[i].data + 8, NULL);
    }
    if (ddata->protver < 15) {
      warnx("hw: unsupported protocol version %f", ddata->protver);
      free(ddata);
      return NULL;
    }

    ddata->port = prtid_INVALID; /* invalid: set below */
    ddata->period = 0.;
  } while(!ddata);

  /* determine the current port - this is cumbersome
   * First set the rate of a message to 255, read back the settings on all
   * ports to find the non-zero one ... */
  ublox_msg(sizeof(struct ublox_cfg_msg)) cfg;

  /* set 255 rate for TIM-VRFY (supported since protocol 15) */
  cfg.msg.data->cfg_msg = (struct ublox_cfg_msg){
    .msgclass = mcid_TIM_VRFY >> 8,
    .msgid = mcid_TIM_VRFY & 0xff,
    .rate = { 255 }
  };
  s = ublox_ack_msg(chan, &cfg.msg, mcid_CFG_MSG, 3, 500);
  if (s) goto failed;

  /* read settings for all ports (sending only 2 bytes) */
  s = ublox_ack_msg(chan, &cfg.msg, mcid_CFG_MSG, 2, 500);
  if (s) goto failed;

  /* find port */
  for (i=0; i < sizeof(cfg.msg.data->cfg_msg.rate); i++)
    if (cfg.msg.data->cfg_msg.rate[i] == 255) { ddata->port = i; break; }

  /* set back to 0 rate for current port (sending only 3 bytes) */
  cfg.msg.data->cfg_msg.rate[0] = 0;
  s = ublox_ack_msg(chan, &cfg.msg, mcid_CFG_MSG, 3, 500);
  if (s) goto failed;

  /* read settings for all ports (sending only 2 bytes) */
  s = ublox_ack_msg(chan, &cfg.msg, mcid_CFG_MSG, 2, 500);
  if (s) goto failed;

  /* check port is ok */
  if (ddata->port == prtid_INVALID) goto failed;
  if (cfg.msg.data->cfg_msg.rate[ddata->port] != 0) goto failed;

  return ddata;

failed:
  if (ddata) free(ddata);
  warnx("detected %s protocol at %dbps, but cannot determine port",
        ublox_driver.protocol, chan->baud);
  return NULL;
}


/* --- ublox_configure ----------------------------------------------------- */

static const struct {
  enum ublox_msgcid msgcid; uint8_t rate;
} default_cfg_msg[] = {
  /* disable default NMEA messages */
  { mcid_NMEA_GGA, 0 },
  { mcid_NMEA_GLL, 0 },
  { mcid_NMEA_GSA, 0 },
  { mcid_NMEA_GSV, 0 },
  { mcid_NMEA_RMC, 0 },
  { mcid_NMEA_VTG, 0 },

  /* disable RTCM messages */
  { mcid_RTCM_1005, 0 },
  { mcid_RTCM_1074, 0 },
  { mcid_RTCM_1077, 0 },
  { mcid_RTCM_1084, 0 },
  { mcid_RTCM_1087, 0 },
  { mcid_RTCM_1124, 0 },
  { mcid_RTCM_1127, 0 },
  { mcid_RTCM_1230, 0 },

  /* enable UBX messages */
  { mcid_NAV_EOE, 1 },
  { mcid_NAV_HPPPOSLLH, 1 },
  { mcid_NAV_PVT, 1 },
  { mcid_RXM_RTCM, 1 },
  { mcid_MON_MSGPP, 7 },
  { mcid_MON_TXBUF, 5 }
};

static int
ublox_configure(struct gps_device_s *dev)
{
  struct ublox_driver_data_s *ddata = dev->data;
  unsigned int i;
  size_t msgsz;
  int s;

  /* port configuration - resend the message thrice as it is sometimes
   * not acked if the baud rate is changed on UART */
  for(i = 0; i < 3; i++) {
    s = ublox_ctl(dev, GPSCTL_BAUD, dev->channel.baud);
    if (s) return s;
    s = ublox_ack_msg(&dev->channel, NULL, mcid_CFG_PRT, 0, 500);
    if (!s) break;
  }
  if (s) {
    warnx("u-blox port configuration failed");
    return s;
  }


  /* periodic messages */
  for(i = 0; i < sizeof(default_cfg_msg)/sizeof(default_cfg_msg[0]); i++) {
    ublox_msg(sizeof(struct ublox_cfg_msg)) rate = {};

    rate.msg.data->cfg_msg.msgclass = default_cfg_msg[i].msgcid >> 8;
    rate.msg.data->cfg_msg.msgid = default_cfg_msg[i].msgcid & 0xff;
    rate.msg.data->cfg_msg.rate[0] = default_cfg_msg[i].rate;

    if (default_cfg_msg[i].rate) {
      /* enable on current port */
      s = ublox_ack_msg(&dev->channel, &rate.msg, mcid_CFG_MSG, 3, 500);
    } else
      /* disable on all ports - otherwise quickly gives txbuf errors */
      s = ublox_ack_msg(&dev->channel, &rate.msg, mcid_CFG_MSG,
                        sizeof(struct ublox_cfg_msg), 500);
    if (s) {
      warnx("u-blox message 0x%04x failed", default_cfg_msg[i].msgcid);
      return s;
    }
  }


  /* information messages */
  ublox_msg(sizeof(struct ublox_cfg_inf)) inf = {};

  /* NMEA */
  inf.msg.data->cfg_inf.proto = 1;
  s = ublox_ack_msg(&dev->channel, &inf.msg, mcid_CFG_INF, 1, 500);
  if (s) return s;
  inf.msg.data->cfg_inf.mask[ddata->port] = 0;	/* none */
  s = ublox_ack_msg(
    &dev->channel, &inf.msg, mcid_CFG_INF, sizeof(inf.payload), 500);
  if (s) return s;

  /* UBX */
  inf.msg.data->cfg_inf.proto = 0;
  s = ublox_ack_msg(&dev->channel, &inf.msg, mcid_CFG_INF, 1, 500);
  if (s) return s;
  inf.msg.data->cfg_inf.mask[ddata->port] = 7;	/* NOTICE, WARN, ERR */
  s = ublox_ack_msg(
    &dev->channel, &inf.msg, mcid_CFG_INF, sizeof(inf.payload), 500);
  if (s) return s;


  /* power mode (not allowed on USB: always full power) */
  if (ddata->port != prtid_USB) {
    ublox_msg(sizeof(struct ublox_cfg_pms)) pms = {0};

    pms.msg.data->cfg_pms.powersetupvalue = 0;	/* full power */
    s = ublox_ack_msg(
      &dev->channel, &pms.msg, mcid_CFG_PMS, sizeof(pms.payload), 500);
    if (s) return s;
  }


  /* differential mode */
  ublox_msg(sizeof(struct ublox_cfg_dgnss)) dgnss = {};
  dgnss.msg.data->cfg_dgnss.dgnssMode = 3; /* RTK fixed if possible */
  s = ublox_ack_msg(
    &dev->channel, &dgnss.msg, mcid_CFG_DGNSS, sizeof(dgnss.payload), 500);
  if (s) return s;


  /* computation rate */
  ublox_msg(sizeof(struct ublox_cfg_rate)) rate = {};

  rate.msg.data->cfg_rate.measrate = 50;	/* 20Hz */
  rate.msg.data->cfg_rate.navrate = 1;		/* on every measurement */
  rate.msg.data->cfg_rate.timeref = 0;		/* UTC time sync */
  s = ublox_ack_msg(
    &dev->channel, &rate.msg, mcid_CFG_RATE, sizeof(struct ublox_cfg_rate),
    500);
  if (s) return s;

  ddata->period = 0.05;
  switch(ddata->port) {
    case prtid_UART1: /* delay with 8N1 UART */
      msgsz =
        sizeof(ublox_msg(sizeof(struct ublox_nav_pvt))) +
        sizeof(ublox_msg(sizeof(struct ublox_nav_hppposllh))) +
        sizeof(ublox_msg(sizeof(struct ublox_nav_eoe)));
      ddata->tdelay = msgsz * 10./dev->channel.baud;
      break;
    case prtid_USB: /* assume 1ms delay on USB (frame) */
      ddata->tdelay = 1e-3;
      break;
  }

  return 0;
}


/* --- ublox_unconfigure --------------------------------------------------- */

static int
ublox_unconfigure(struct gps_device_s *dev)
{
  ublox_msg(sizeof(struct ublox_cfg_msg)) rate = {};
  unsigned int i;

  /* periodic messages */
  for(i = 0; i < sizeof(default_cfg_msg)/sizeof(default_cfg_msg[0]); i++) {
    rate.msg.data->cfg_msg.msgclass = default_cfg_msg[i].msgcid >> 8;
    rate.msg.data->cfg_msg.msgid = default_cfg_msg[i].msgcid & 0xff;
    rate.msg.data->cfg_msg.rate[0] = 0;

    ublox_ack_msg(&dev->channel, &rate.msg, mcid_CFG_MSG, 3, 500);
  }

  return 0;
}


/* --- ublox_decode -------------------------------------------------------- */

static int
ublox_decode(struct gps_device_s *dev, struct gps_solution_s *solution,
             sequence_octet *rtk_data)
{
  (void)rtk_data;
  struct ublox_driver_data_s *ddata = dev->data;
  const struct ublox_msg *msg;
  const uint8_t *rtcm;
  uint16_t rtcmid;
  struct timeval tv;
  size_t len;

  while(1) {
    gettimeofday(&tv, NULL);
    len = ublox_recv_msg(&dev->channel, NULL, &msg, &rtcm, &rtcmid);
    if (len == 0) break;

    if (rtcm) {
      if (rtk_data->_maximum < rtk_data->_length + len)
        if (genom_sequence_reserve(rtk_data, rtk_data->_length + len))
          return (errno = ENOMEM);
      memcpy(&rtk_data->_buffer[rtk_data->_length], rtcm, len);
      rtk_data->_length += len;

      /* increment message counters */
      rtcmidseq_insert(&ddata->rtcmstats.sent, rtcmid, i,
                       ddata->rtcmstats.sent._buffer[i].out++);
      ddata->rtcmstats.outbytes += len;
      ddata->rtcmstats.outpkts++;
    }

    if (!msg) continue;
    switch(ublox_msgcid16(&msg->hdr)) {
      case mcid_RXM_RTCM: {
        const struct ublox_rxm_rtcm *rxm = &msg->data->rxm_rtcm;

        if (rxm->flags & 1)
          warnx("RTCM message %d had CRC failures", rxm->msgtype);
        break;
      }

      case mcid_MON_MSGPP: {
        const struct ublox_mon_msgpp *pp = &msg->data->mon_msgpp;

        if (ddata->skipped &&
            ddata->skipped != pp->skipped[ddata->port])
          warnx("Hardware skipped %d bytes",
                pp->skipped[ddata->port] - ddata->skipped);

        ddata->skipped = pp->skipped[ddata->port];
        break;
      }

      case mcid_MON_TXBUF: {
        const struct ublox_mon_txbuf *tx = &msg->data->mon_txbuf;
        int i;

        for (i = 0; i < 6; i++) {
          if (tx->usage[i] > 90 || tx->peakusage[i] > 90)
            warnx("hw transmit buffer %d: %d bytes, %d%% usage, %d%% peak",
                  i, tx->pending[i], tx->usage[i], tx->peakusage[i]);
        }
        break;
      }

      case mcid_NAV_HPPPOSLLH:
        ddata->hppposllh = msg->data->nav_hppposllh;
        break;

      case mcid_NAV_PVT:
        ddata->pvt = msg->data->nav_pvt;
        break;

      case mcid_NAV_EOE:
        ublox_solution(msg->data->nav_eoe.itow, &tv, ddata, solution);
        return 0 /* publish now */;

      default:
        warnx("unexpected message 0x%04x", ublox_msgcid16(&msg->hdr));
    }
  }

  return EAGAIN;
}

static void
ublox_solution(uint32_t itow, const struct timeval *tv,
               struct ublox_driver_data_s *ddata,
               struct gps_solution_s *solution)
{
  const struct ublox_nav_pvt *pvt = &ddata->pvt;
  const struct ublox_nav_hppposllh *hppposllh = &ddata->hppposllh;
  or_time_ts utc, ts;

  if (itow != pvt->itow) /* no NAV-PVT for this epoch */ return;
  if (!(pvt->valid & 0x7)) /* invalid time data */ return;

  utc = gpsconv_utcmktime(
    &(struct tm){
        .tm_sec = pvt->sec,
        .tm_min = pvt->min,
        .tm_hour = pvt->hour,
        .tm_mday = pvt->day,
        .tm_mon = pvt->month - 1,
        .tm_year = pvt->year - 1900
    }, pvt->nano);
  ts = gpsconv_utcts(utc, *tv, ddata->period, ddata->tdelay);

  /* default solution from NAV-PVT */
  *solution = (gps_solution_s){
    .ts = ts,
    .utc = utc,
    .fix = gps_fix_none /* by default, but updated below */,
    .sats = pvt->numsv,

    .llh = {
      .latitude = pvt->lat * 1e-7,
      .longitude = pvt->lon * 1e-7,
      .height = pvt->height * 1e-3
    },
    .pos_sigma = {
      .east = pvt->hacc * 1e-3,
      .north = pvt->hacc * 1e-3,
      .up = pvt->vacc * 1e-3
    },

    .vel =  {
      .east = pvt->vele * 1e-3,
      .north = pvt->veln * 1e-3,
      .up = - pvt->veld * 1e-3
    },
    .vel_sigma = {
      .east = pvt->sacc * 1e-3,
      .north = pvt->sacc * 1e-3,
      .up = pvt->sacc * 1e-3
    }
  };

  /* fix type */
  switch(pvt->fixtype) {
    case 3: /* 3D fix */
    case 4: /* 3D fix + dead reckoning */
      if (!(pvt->flags & 0x1)) break; /* invalid fix */

      if (pvt->flags & 0x80) solution->fix = gps_fix_rtk;
      else if (pvt->flags & 0x40) solution->fix = gps_fix_rtk_float;
      else if (pvt->flags & 0x2) solution->fix = gps_fix_diff;
      else solution->fix = gps_fix_single;

      break;

    case 5: /* fixed position */
      solution->fix = gps_fix_fixed;
      break;
  }

  /* high precision position */
  if (itow == hppposllh->itow) {
    solution->llh = (gps_llh_s){
      .latitude = hppposllh->lat * 1e-7 + hppposllh->lathp * 1e-9,
      .longitude = hppposllh->lon * 1e-7 + hppposllh->lonhp * 1e-9,
      .height = hppposllh->height * 1e-3 + hppposllh->heighthp * 1e-4,
    };

    solution->pos_sigma = (gps_enu_s){
      .east = hppposllh->hacc * 1e-4,
      .north = hppposllh->hacc * 1e-4,
      .up = hppposllh->vacc * 1e-4
    };
  }

  /* ENU position according to configured reference */
  gpsconv_llhenu(&solution->llh, &solution->pos);
}


/* --- ublox_ctl ----------------------------------------------------------- */

static int
ublox_ctl(struct gps_device_s *dev, enum gps_ctl_s request, ...)
{
  struct ublox_driver_data_s *ddata = dev->data;
  va_list args;
  int s = 0;

  va_start(args, request);
  switch(request) {
    case GPSCTL_BAUD: /* set baud rate */ {

      uint32_t baud = va_arg(args, uint32_t);

      ublox_msg(sizeof(struct ublox_cfg_prt)) prt = {};
      prt.msg.data->cfg_prt.portid = ddata->port;
      prt.msg.data->cfg_prt.txready = 0;		/* tx pin disabled */
      switch(ddata->port) {
        case prtid_UART1:
          prt.msg.data->cfg_prt.uart.mode =
            (3<<6)|(4<<9);				/* 8N1 */
          prt.msg.data->cfg_prt.uart.baud = baud;	/* baud */
          prt.msg.data->cfg_prt.uart.inproto =
            (1<<0)|(1<<2)|(1<<5);			/* UBX/RTCM2/RTCM3 */
          prt.msg.data->cfg_prt.uart.outproto =
            (1<<0)|(1<<5);				/* UBX/RTCM3 */
          prt.msg.data->cfg_prt.uart.flags = 0;		/* 1.5s timeout */
          break;

        case prtid_USB:
          prt.msg.data->cfg_prt.usb.inproto =
            (1<<0)|(1<<2)|(1<<5);			/* UBX/RTCM2/RTCM3 */
          prt.msg.data->cfg_prt.usb.outproto =
            (1<<0)|(1<<5);				/* UBX/RTCM3 */
          break;

        default: s = 1; break;
      }
      if (s) break;

      s = ublox_send_msg(
        &dev->channel, &prt.msg, mcid_CFG_PRT, sizeof(struct ublox_cfg_prt));
      break;
    }

    case GPSCTL_UNFIX: /* unset fixed position */ {
      ublox_msg(sizeof(struct ublox_cfg_tmode3)) tmode3 = {};
      tmode3.msg.data->cfg_tmode3.flags = 0;

      s = ublox_ack_msg(
        &dev->channel, &tmode3.msg,
        mcid_CFG_TMODE3, sizeof(struct ublox_cfg_tmode3), 500);
      break;
    }

    case GPSCTL_FIX: /* set fixed position */ {
      struct gps_llh_s fix = *va_arg(args, const struct gps_llh_s *);
      struct gps_enu_s sigma = *va_arg(args, const struct gps_enu_s *);

      ublox_msg(sizeof(struct ublox_cfg_tmode3)) tmode3 = {};
      tmode3.msg.data->cfg_tmode3.flags = (1<<8)|2 /* lla / fixed mode */;
      tmode3.msg.data->cfg_tmode3.lat = fix.latitude * 1e7;
      tmode3.msg.data->cfg_tmode3.lon = fix.longitude * 1e7;
      tmode3.msg.data->cfg_tmode3.height = fix.height * 1e2;
      tmode3.msg.data->cfg_tmode3.lathp =
        (fix.latitude * 1e7 - tmode3.msg.data->cfg_tmode3.lat) * 1e2;
      tmode3.msg.data->cfg_tmode3.lonhp =
        (fix.longitude * 1e7 - tmode3.msg.data->cfg_tmode3.lon) * 1e2;
      tmode3.msg.data->cfg_tmode3.heighthp =
        (fix.height * 1e2 - tmode3.msg.data->cfg_tmode3.height) * 1e2;
      if (sigma.east >= sigma.north && sigma.east >= sigma.up)
        tmode3.msg.data->cfg_tmode3.fixedposacc = 2 * sigma.east * 1e4;
      else if (sigma.north >= sigma.east && sigma.north >= sigma.up)
        tmode3.msg.data->cfg_tmode3.fixedposacc = 2 * sigma.north * 1e4;
      else if (sigma.up >= sigma.east && sigma.up >= sigma.north)
        tmode3.msg.data->cfg_tmode3.fixedposacc = 2 * sigma.up * 1e4;

      s = ublox_ack_msg(
        &dev->channel, &tmode3.msg,
        mcid_CFG_TMODE3, sizeof(struct ublox_cfg_tmode3), 500);
      break;
    }

    case GPSCTL_INRTCM: /* feed RTCM data */ {
      uint16_t type = va_arg(args, unsigned int);
      uint8_t *data = va_arg(args, uint8_t *);
      size_t len = va_arg(args, size_t);

      if (len == 0) { s = 0; break; }

      switch (type) {
        /* RTCM2 */
        case 1: case 2:
        case 3: case 9:
          break;

          /* RTCM3 */
        case 1005: case 1006: case 1007: /* base station */
        case 1009: case 1010: case 1011: case 1012: /* observations */
        case 1074: case 1075: case 1077: /* GPS MSM 4, 5, 7 */
        case 1084: case 1085: case 1087: /* GLONASS MSM 4, 5, 7 */
        case 1124: case 1125: case 1127: /* Beidou MSM 4, 5, 7 */
        case 1230: /* GLONASS biases */
          break;

        default: s = 1;
      }
      if (!s) s = gpstty_send(&dev->channel, data, len);

      /* increment message counters */
      if (s) {
        ddata->rtcmstats.dropbytes += len;
        ddata->rtcmstats.droppkts++;
      } else {
        rtcmidseq_insert(&ddata->rtcmstats.recv, type, i,
                         ddata->rtcmstats.recv._buffer[i].in++);
        ddata->rtcmstats.inbytes += len;
        ddata->rtcmstats.inpkts++;
      }
      break;
    }

    case GPSCTL_OUTRTCM: /* stream RTCM data */ {
      uint16_t rtcmid = va_arg(args, unsigned int);
      double period = va_arg(args, double);
      double offset = va_arg(args, double);

      ublox_msg(sizeof(struct ublox_cfg_msg)) rate = {};
      enum ublox_msgcid mcid;

      offset = 0.;
      switch(rtcmid) {
        case 1005: mcid = mcid_RTCM_1005; break;
        case 1074: mcid = mcid_RTCM_1074; break;
        case 1077: mcid = mcid_RTCM_1077; break;
        case 1084: mcid = mcid_RTCM_1084; break;
        case 1087: mcid = mcid_RTCM_1087; break;
        case 1124: mcid = mcid_RTCM_1124; break;
        case 1127: mcid = mcid_RTCM_1127; break;
        case 1230: mcid = mcid_RTCM_1230; break;

        default: return 1;
      }

      /* set on current port (sending only 3 bytes) */
      rate.msg.data->cfg_msg.msgclass = mcid >> 8;
      rate.msg.data->cfg_msg.msgid = mcid & 0xff;
      rate.msg.data->cfg_msg.rate[0] =
        ddata->period > 0. ? period/ddata->period : 0.;
      s = ublox_ack_msg(&dev->channel, &rate.msg, mcid_CFG_MSG, 3, 500);
      if (s) break;

      if (period > 0.) {
        /* append to the list of sent messages */
        rtcmidseq_insert(
          &ddata->rtcmstats.sent, rtcmid, i,
          ddata->rtcmstats.sent._buffer[i].period =
            rate.msg.data->cfg_msg.rate[0] * ddata->period;
          ddata->rtcmstats.sent._buffer[i].offset = offset);
      } else {
        /* remove from the list of sent and recv messages */
        rtcmidseq_rm(&ddata->rtcmstats.sent, rtcmid);
        rtcmidseq_rm(&ddata->rtcmstats.recv, rtcmid);
      }
      break;
    }

    case GPSCTL_Q_PARAM: /* query model */ {
      gps_param_s *param = va_arg(args, gps_param_s *);

      snprintf(param->model, sizeof(param->model), "%s", ddata->model);
      param->rate = ddata->period > 0. ? 1./ddata->period : 0.;
      break;
    }

    case GPSCTL_Q_RTCM: /* RTCM stats */ {
      s = gpsrtk_copy_rtcm_stats(
        va_arg(args, gps_rtcm_stats_s *), &ddata->rtcmstats);
      break;
    }

    default: errno = ENOENT; s = 1;
  }
  va_end(args);

  return s;
}


/* --- ublox_ack_msg ------------------------------------------------------- */

static int
ublox_ack_msg(struct gps_channel_s *chan, struct ublox_msg *msg,
              enum ublox_msgcid mcid, uint16_t len, int timeout)
{
  const struct ublox_msg *ack;
  struct timeval deadline;
  int s;

  gettimeofday(&deadline, NULL);
  deadline.tv_usec += timeout * 1000;
  if (deadline.tv_usec > 1000000) {
    deadline.tv_sec += deadline.tv_usec / 1000000;
    deadline.tv_usec = deadline.tv_usec % 1000000;
  }

  if (msg) {
    s = ublox_send_msg(chan, msg, mcid, len);
    if (s) return s;
  }

  /* ignore all messages until corresponding ACK */
  while(1) {
    if (ublox_recv_msg(chan, &deadline, &ack, NULL, NULL) == 0) {
      warnx("timeout for message 0x%04x", mcid); return -1;
    }

    /* if there is a response, save it back to the message */
    if (ublox_msgcid16(&ack->hdr) == mcid)
      memcpy(msg, ack, sizeof(ublox_msg(ack->hdr.msglen)));

    if (ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK &&
        ublox_msgcid16(&ack->hdr) != mcid_ACK_NAK) continue;
    if (ublox_msgcid16(&ack->data->ack) == mcid)
      break;
  }

  return ublox_msgcid16(&ack->hdr) != mcid_ACK_ACK;
}


/* --- ublox_recv_msg ------------------------------------------------------ */

static size_t
ublox_recv_msg(struct gps_channel_s *chan, const struct timeval *deadline,
               const struct ublox_msg **ubx,
               const uint8_t **rtcm, uint16_t *rtcmid)
{
  static const struct ublox_hdr expect = {
    .sync1 = 0xb5, .sync2 = 0x62
  };

  const struct ublox_hdr *hdr;
  const void *raw;
  size_t len, req;
  ssize_t n;

  if (ubx) *ubx = NULL;
  if (rtcm) *rtcm = NULL;

  do {
    /* peek sizeof(hdr) data */
    hdr = raw = gpstty_data(chan, 0, &len);
    if (len < sizeof(*hdr)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break; /* err or not enough data */
      continue;
    }

    /* check raw RTCM3 (needs 5 bytes, less than sizeof(hdr)) */
    if (gpsrtk_is_rtcm(raw, rtcmid, &req) == 3) {
      for(; len < req; gpstty_data(chan, 0, &len)) {
        n = gpstty_recv(chan, deadline);
        if (n <= 0) break;
      }
      if (len < req) break; /* err or not enough data */

      raw = gpstty_data(chan, req, &len);
      if (!rtcm) continue; /* skip unwanted rtcm data */

      *rtcm = raw;
      return req;
    }

    /* match sync bytes */
    if (hdr->sync1 != expect.sync1 || hdr->sync2 != expect.sync2) {
      gpstty_data(chan, 1, NULL); /* skip 1 byte */
      continue;
    }

    /* get enough data for the whole message */
    const ublox_msg(hdr->msglen) *msg;

    for(req = sizeof(*msg); len < req; gpstty_data(chan, 0, &len)) {
      n = gpstty_recv(chan, deadline);
      if (n <= 0) break;
    }

    msg = gpstty_data(chan, 0, &len);
    if (len < req) break; /* err or not enough data */

    /* check crc */
    if (msg->cksum != ublox_cksum(&msg->msg.hdr)) {
      warnx("u-blox message 0x%02x%02x with bad checksum",
            msg->msg.hdr.msgclass, msg->msg.hdr.msgid);

      gpstty_data(chan, 2, NULL); /* skip 2 bytes */
      continue;
    }

    gpstty_data(chan, req, NULL);

    /* skip and display information messages */
    switch(ublox_msgcid16(&msg->msg.hdr)) {
      case mcid_INF_ERROR:
        warnx("hw error: %.*s", msg->msg.hdr.msglen, &msg->msg.data->inf.str);
        continue;
      case mcid_INF_WARN:
        warnx("hw warning: %.*s", msg->msg.hdr.msglen, &msg->msg.data->inf.str);
        continue;
      case mcid_INF_NOTICE:
        warnx("hw: %.*s", msg->msg.hdr.msglen, &msg->msg.data->inf.str);
        continue;
      case mcid_INF_TEST:
        warnx("hw test: %.*s", msg->msg.hdr.msglen, &msg->msg.data->inf.str);
        continue;
      case mcid_INF_DEBUG:
        warnx("hw debug: %.*s", msg->msg.hdr.msglen, &msg->msg.data->inf.str);
        continue;
    }

    if (!ubx) continue;  /* skip unwanted message */

    *ubx = &msg->msg;
    return req;

  } while(1);

  return 0;
}


/* --- ublox_send_msg ------------------------------------------------------ */

static int
ublox_send_msg(struct gps_channel_s *chan, struct ublox_msg *msg,
               enum ublox_msgcid mcid, uint16_t len)
{
  msg->hdr = (struct ublox_hdr){
    .sync1 = 0xb5, .sync2 = 0x62,
    .msgclass = mcid >> 8,
    .msgid = mcid & 0xff,
    .msglen = len
  };
  ((ublox_msg(len) *)msg)->cksum = ublox_cksum(&msg->hdr);

  return gpstty_send(chan, msg, sizeof(ublox_msg(len)));
}


/* --- ublox_cksum --------------------------------------------------------- */

static uint16_t
ublox_cksum(const struct ublox_hdr *hdr)
{
  uint16_t len = hdr->msglen;
  uint8_t cksum_a, cksum_b;
  const uint8_t *p;

  cksum_a  = hdr->msgclass;      cksum_b  = cksum_a;
  cksum_a += hdr->msgid;         cksum_b += cksum_a;
  cksum_a += hdr->msglen & 0xff; cksum_b += cksum_a;
  cksum_a += hdr->msglen >> 8;   cksum_b += cksum_a;

  for(p = hdr->payload; len--; p++) {
    cksum_a += *p;               cksum_b += cksum_a;
  }

  return (cksum_b << 8) | cksum_a;
}
